;;; -*- Mode: Lisp; Package: CHAOS-INTERNALS; Base: 10 -*-

(herald chaos)

(eval-when (eval compile)
  (or (get 'pktdefs 'version)
      (load "nil$chaos:pktdefs"))
  (or (get 'syscalls 'version)
      (load "nil$chaos:syscalls")))


(defresource packet ()
  :constructor (make-packet))

(defun make-packet ()
  (make-bits (*& 8 pkt$k_pktmax)))

(eval-when (compile eval load)
  (defmacro with-packet ((name) &body bod)
    `(using-resource (packet ,name) ,@bod))
  (defmacro with-iosb ((name) &body bod)
    `(using-resource (si:bit-buffer ,name (*& 8 iosb$k_iosbmax)) ,@bod))
  (defmacro with-csb ((name) &body bod)	; connection state buffer
    `(using-resource (si:bit-buffer ,name (*& 8 conn$k_connmax)) ,@bod))
  (defmacro with-timeout ((efn-name timeout reqidt) &body bod)
    (let ((timbuf-name (gensym))
	  (req-name (gensym)))
      `(using-resource (si:bit-buffer ,timbuf-name 64.)
	 (let ((,req-name (si:address-of ,reqidt)))
	   (unwind-protect
	    (progn
	     (unless (zerop ,timeout)
	       ($setimr efn ,efn-name
			daytim (to-delta-time ,timeout ':sixtieths
					      ,timbuf-name)
			reqidt ,req-name))
	     ,@bod)
	    (when (not (zerop ,timeout))
	      ($cantim reqidt ,req-name)))))))

  (defmacro with-connection ((conn-name conn-form)
			     &body body)
    `(let ((,conn-name nil))
       (unwind-protect
	(progn
	 (setq ,conn-name ,conn-form)
	 ,@body)
	(and ,conn-name (chaos:close-conn ,conn-name)))))
  )

(globalize 'packet 'chaos)
(globalize 'make-packet 'chaos)
(globalize 'with-packet 'chaos)
(globalize 'with-connection 'chaos)
(globalize 'with-iosb 'chaos)

;;;; definition of the connection

(defflavor connection
	   ((vms-channel 0)
	    (vms-efn 0)			; the event flag to use
	    (physical-name nil)		; the physical name of the CH unit
	    (foreign-host nil)		; if we ever find out...
	    (rfc? nil)			; remember so we can print pretty
	    (contact-name-used nil)	; this, too
	    )
	   ()
  (:gettable-instance-variables physical-name foreign-host))


(defresource connection ()
  :constructor (make-instance 'connection)
  :initializer (send object ':vms-initialization)
  :deinitializer (send object ':vms-cleanup))

(defconstant *chaos-state-alist*
  '((#.conn$k_closed . chaos:closed-state)
    (#.conn$k_rfcrcv . chaos:rfc-recieved-state)
    (#.conn$k_rfcsnt . chaos:rfc-sent-state)
    (#.conn$k_open . chaos:open-state)
    (#.conn$k_los . chaos:los-recieved-state)
    (#.conn$k_incom . chaos:host-down-state)
    (#.conn$k_new . chaos:inactive-state)
    (#.conn$k_lsn . chaos:listening-state)))


;;;; connection methods

;;; :vms-initialization - performs initial allocation and assignment of
;;; the VMS channel.

(defmethod (connection :vms-initialization) ()
  (let ((phyname nil)
	stat)
    (if (zerop vms-efn)
	(progn (setq stat (si:%allocate-event-flag))
	       (when (< stat 0)
		 (ferror ':vms-error "Couldn't allocate an event flag"))
	       (setq vms-efn stat)))
    (setq phyname (si:%allocate-system-device "_CH"))
    (when (numberp phyname)
      (si:loschk "allocate" phyname))
    (setq stat ($assign devnam phyname chan vms-channel))
    (if (si:rms-losing stat)
	(progn (si:%deallocate-system-device phyname)	; release for use
	       (si:loschk "assign" stat)))	; signal error
    (setq foreign-host nil)		; don't want spurious stuff here
    (setq rfc? nil contact-name-used nil)	; and this stuff
    (setq physical-name phyname)))

(defmethod (connection :vms-cleanup) ()
  (if (zerop vms-channel) ()
      (progn ($dassgn chan vms-channel)
	     (si:%deallocate-system-device physical-name)
	     (si:%free-event-flag vms-efn)
	     (setq vms-channel 0 physical-name nil vms-efn 0))))


(defmethod (connection :print-self) (stream level slash?)
  (if rfc?
      (format (or stream si:standard-output)
	      "#<Chaos connection to ~A/~A ~16r>"
	      (host-name foreign-host)
	      contact-name-used
	      (si:address-of self))
      (if foreign-host
	  (format (or stream si:standard-output)
		  "#<Chaos connection from ~A to ~A server ~16r"
		  (host-name foreign-host)
		  contact-name-used
		  (si:address-of self))
	  (format (or stream si:standard-output)
		  "#<Chaos connection from unknown host to ~A server ~16r"
		  contact-name-used (si:address-of self)))))

(defmethod (connection :remove-self) ()
  (cond ((zerop vms-channel) nil)
	(t (deallocate-resource 'connection self))))

(defmethod (connection :send-buffer) (&key data &optional (opcode pkt$k_data)
					   (length 0 l-p))
  (let ((datlen (buffer-size data)))
    (if (and l-p (< length datlen)) (setq datlen length))
    (if (> datlen pkt$k_maxdata)
	(ferror ':chaos-lossage "Data too large in :SEND-PACKET: ~D > 488"
		datlen))
    (if (zerop vms-channel)
	(ferror ':chaos-lossage "Connection not initialized, :SEND-PACKET"))
    (with-packet (pkt)
      (copy-data-to-packet data pkt datlen)
      (setf (pkt$b_opcode pkt) opcode)
      (setf (pkt$w_nbytes pkt) datlen)
      (with-iosb (iosb)
	(si:loschk "qio"
		   ($pktio efn vms-efn
			   chan vms-channel
			   func io$_writepkt
			   iosb iosb
			   pkt pkt
			   pktlen (+ datlen pkt$k_hdrlen)))
	(si:loschk "packet i/o status" (iosb$w_io_status iosb))))))


(defmethod (connection :send-pkt) (&key pkt &optional (opcode pkt$k_data))
  (if (zerop vms-channel)
      (ferror ':chaos-lossage "Connection not initialized, :SEND-PACKET"))
  (setf (pkt$b_opcode pkt) opcode)
  (with-iosb (iosb)
    (si:loschk "qio"
	       ($pktio efn vms-efn
		       chan vms-channel
		       func io$_writepkt
		       iosb iosb
		       pkt pkt
		       pktlen (+ (pkt$w_nbytes pkt) pkt$k_hdrlen)))
    (si:loschk "packet i/o status" (iosb$w_io_status iosb))))

(defmethod (connection :get-packet) (&optional &key pkt (timeout nil))
  (if (zerop vms-channel)
      (ferror ':chaos-lossage "Connection not initialized, :GET-PACKET"))
  (when (and (not (null timeout))
	     (zerop timeout))
    (setq timeout nil))
  (let ((ipkt nil) (ok nil))
    (unwind-protect
     (with-csb (buff)
       (if pkt (setq ipkt pkt) (setq ipkt (allocate-resource 'packet)))
       (loop for state = (send self ':state ':buffer buff)
	     for once = nil then t
	     with itimeout = (if (null timeout) 7200 timeout)
	     until (or (and timeout once) ok)
	     do (if (and (or (= state conn$k_closed)
			     (= state conn$k_los)
			     (= state conn$k_incom))
			 (zerop (conn$w_rcv_available buff)))
		    (ferror ':chaos-lossage
			    "Attempt to read on dead connection. ~S~%~
			        state = ~S, rcv avail = ~D"
			    self
			    (state-sym state)
			    (conn$w_rcv_available buff)))
	     (setq ok (send self ':get-packet-internal ipkt itimeout))))
     (unless ok (when (and (null pkt) ipkt)
		  (deallocate-resource 'packet ipkt))))
    (when ok
      (unless (eq foreign-host (pkt$w_src_host ipkt))
	(setq foreign-host (pkt$w_src_host ipkt))))
    (if ok ipkt nil)))

(defmethod (connection :get-packet-internal) (pkt timeout)
  (with-timeout (vms-efn timeout self)
    (with-iosb (iosb)
      (si:loschk "qio"
		 ($pktio efn vms-efn
			 chan vms-channel
			 func io$_readpkt
			 iosb iosb
			 pkt pkt
			 pktlen pkt$k_pktmax))
      (si:loschk "pkt i/o status" (iosb$w_io_status iosb))
      (if (zerop (iosb$w_io_status iosb))	; timed out!
	  (progn (si:%cancel-system-io vms-channel) nil)
	  pkt))))

(defmethod (connection :state) (&optional &key (buffer nil))
  (if (zerop vms-channel)
      (ferror ':chaos-lossage "VMS channel not open for connection ~S" self))
  (with-csb (stbuf)
    (unless buffer (setq buffer stbuf))
    (with-iosb (iosb)
      (si:loschk "state qio" ($pktio efn vms-efn
				     chan vms-channel
				     func io$_readstatus
				     iosb iosb
				     pkt buffer
				     pktlen conn$c_connmax))
      (si:loschk "state qio return status" (iosb$w_io_status iosb))
      (conn$w_state buffer))))

(defmethod (connection :may-transmit) ()
  (with-csb (buff)
    (and (= (send self ':state ':buffer buff) conn$k_open)
	 (not (zerop (conn$w_xmit_available buff))))))

(defmethod (connection :data-available) ()
  (with-csb (buff)
    (send self ':state ':buffer buff)
    (not (zerop (conn$w_rcv_available buff)))))

(defmethod (connection :wait) (&optional &key (new-state nil np)
					      (old-state nil op)
					      (timeout 0))
  (cond ((not (or np op)) nil)
	((and op np (= new-state old-state)) new-state)
	(np (loop for curstate = (send self ':state)
		  until (= curstate new-state)
		  do (send self ':wait ':old-state curstate ':timeout timeout)
		     (when (= (send self ':state) curstate)
		       (return nil))	; timed out...
		  finally (return new-state)))
	(op (with-timeout (vms-efn timeout self)
	      (with-packet (dummy)
		(with-iosb (iosb)
		  (si:loschk "wait qio"
			     ($pktio efn vms-efn
				     chan vms-channel
				     func io$_readstate
				     iosb iosb
				     pkt dummy
				     pktlen pkt$k_pktmax
				     p3 old-state))
		  (if (zerop (iosb$w_io_status iosb))
		      (progn (si:%cancel-system-io vms-channel)
			     nil)
		      (si:loschk "wait i/o status" (iosb$w_io_status iosb))
		      (conn$w_state dummy))))))))

(defmethod (connection :rfc) (&key host contact-name &optional (timeout 0))
  (let ((host-number (host-name-to-number host))
	(contact-name-length (buffer-size contact-name)))
    (with-packet (pkt)
      (setf (pkt$b_opcode pkt) pkt$k_rfc)
      (setf (pkt$w_nbytes pkt) contact-name-length)
      (setf (pkt$w_dest_host pkt) host-number)
      (copy-data-to-packet contact-name pkt contact-name-length)
      (with-iosb (iosb)
	(si:loschk "rfc send qio"
		   ($pktio efn vms-efn
			   chan vms-channel
			   func io$_writepkt
			   iosb iosb
			   pkt pkt
			   pktlen (+ pkt$k_hdrlen contact-name-length)))
	(si:loschk "rfc packet output status" (iosb$w_io_status iosb))))
    (setq foreign-host host-number
	  rfc? t
	  contact-name-used contact-name)
    (send self ':wait ':old-state conn$k_rfcsnt ':timeout timeout)))

(defmethod (connection :listen) (&key contact-name &optional (timeout 0))
  (let ((contact-name-length (buffer-size contact-name)))
    (with-packet (pkt)
      (setf (pkt$b_opcode pkt) pkt$k_lsn)
      (setf (pkt$w_nbytes pkt) contact-name-length)
      (copy-data-to-packet contact-name pkt contact-name-length)
      (with-iosb (iosb)
	(si:loschk "lsn send qio"
		   ($pktio efn vms-efn
			   chan vms-channel
			   func io$_writepkt
			   iosb iosb
			   pkt pkt
			   pktlen (+ pkt$k_hdrlen contact-name-length)))
	(si:loschk "lsn packet output status" (iosb$w_io_status iosb))))
    (setq contact-name-used contact-name rfc? nil)
    (send self ':wait ':new-state conn$k_rfcrcv ':timeout timeout)))

(defmethod (connection :close) (&optional &key (reason ""))
  (with-packet (pkt)
    (set-pkt-string pkt reason)
    (send self ':send-pkt ':pkt pkt ':opcode pkt$k_cls)))


(defmethod (connection :finish) ()
  (with-csb (buff)
    (loop for state = (logand #o177 (send self ':state ':buffer buff))
	  when (= (conn$w_xmit_window buff)
		  (conn$w_xmit_available buff))
	    do (return t)
	  when (not (= state conn$k_open))
	    do (return nil)
	  do (with-timeout (0 6 self)
	       (efn-wait 0)))))

(defun efn-wait (efn)
  (using-resource (si:bit-buffer mask 32)
    (setf (elt mask (\\ efn 32)) 1)
    (si:%wait-for-events efn mask)))

(defmethod (connection :vms-open?) ()
  (not (zerop vms-channel)))

;;;; packet fillers and emptiers

(defun copy-data-to-packet (buffer pkt &optional (length nil) (start 0))
  (unless length (setq length (buffer-size buffer)))
  (if (> (+ length start) pkt$k_maxdata)
      (ferror () "Data too long (~D) to fit in packet starting at ~D"
	      length start))
  (cond ((typep buffer 'bits)
	 (si:bits-replace buffer pkt 0
			  (+ (*& 8 start)
			     pkt$k_hdrlen_bits)
			  (*& 8 length)))
	((typep buffer 'string)
	 (loop for i from 0 below length
	       as j from (+ start pkt$k_hdrlen)
	       do (si:set-a-byte pkt j (char-n buffer i))))
	((typep buffer 'vector)
	 (loop for i from 0 below length
	       as j from (+ start pkt$k_hdrlen)
	       do (si:set-a-byte pkt j (vref buffer i))))
	(t (ferror ':wrong-type-argument
		   "~S bad buffer type, vector, bits or string wanted"
		   (typep buffer))))
  (setf (pkt$w_nbytes pkt) (+ start length)))

(defun copy-data-from-packet (pkt buffer &optional (length nil))
  (unless length (setq length (pkt$w_nbytes pkt)))
  (when (> length (buffer-size buffer))
    (setq length (buffer-size buffer)))
  (cond ((typep buffer 'bits)
	 (si:bits-replace pkt buffer pkt$k_hdrlen_bits 0 (*& 8 length)))
	((typep buffer 'string)
	 (loop for i from 0 below length
	       as j from pkt$k_hdrlen
	       do (rplachar-n buffer i (si:get-a-byte pkt j))))
	((typep buffer 'vector)
	 (loop for i from 0 below length
	       as j from pkt$k_hdrlen
	       do (setf (vref buffer i) (si:get-a-byte pkt j))))
	(t (ferror ':wrong-type-argument
		   "~S bad buffer type, vector, bits or string wanted"
		   (typep buffer)))))

(defun host-name-to-number (host)
  (cond ((numberp host) host)
	((or (symbolp host)
	     (stringp host))
	 (host-address (string-upcase host)))
	(t (ferror ':wrong-type-argument
		   "Can't coerce ~S into a host number" host))))

(defun find-host-name (host)
  (cond ((numberp host) (host-name host))
	((or (symbolp host) (stringp host))
	 (string-upcase host))
	(t (ferror ':wrong-type-argument
		   "Can't coerce ~S into a host name" host))))

(defun state-sym (state)
  (cdr (assq state *chaos-state-alist*)))

(defun state-number (ssym)
  (loop for pair being the elements of *chaos-state-alist*
	when (eq (cdr pair) ssym)
	 do (return (car pair))))

;;; buffer is a quadword bits.
;;; units is a keyword, :seconds, :micros

(defun to-delta-time (value units buffer)
  (let ((nano (units-to-nano (minus value) units)))
    (let ((low-order (remainder nano 1048576))
	  (high-order (quotient nano 1048576)))
      (set-nibble buffer 0 20 low-order)
      (set-nibble buffer 20 20
		  (if (zerop high-order)
		      -1		; fill with ones
		      high-order))
      (set-nibble buffer 40 24 -1))))

(defun units-to-nano (value units)
  (case units
    (:sixtieths (* value 166667))
    (:seconds (* value 10000000))
    (:milliseconds (* value 10000))
    (:micros (* value 10))
    (:minutes (* value 600 1000000))))

(defsubst chk-vms-open (conn op)
  (if (not (send conn ':vms-open?))
      (ferror ':chaos-lossage "Connection ~S not initialized, ~A" conn op)))

(defsubst pkt-data (pkt bytenum)
  (si:get-a-byte pkt (+ bytenum pkt$k_hdrlen)))

(defsubst set-pkt-data (pkt bytenum byte)
  (si:set-a-byte pkt (+ bytenum pkt$k_hdrlen) byte))

(defsubst pkt-data-word (pkt wordnum)
  (nibble pkt (*& 8 (+ (*& wordnum 2) pkt$k_hdrlen)) 16))

(defsubst set-pkt-data-word (pkt wordnum word)
  (set-nibble pkt (*& 8 (+ (*& wordnum 2) pkt$k_hdrlen)) 16
	      word))

(defsubst pkt-opcode (pkt)
  (pkt$b_opcode pkt))

(defsubst pkt-nbytes (pkt)
  (pkt$w_nbytes pkt))

(defun pkt-string (pkt)
  (let ((new (make-string (pkt-nbytes pkt))))
    (copy-data-from-packet pkt new)
    new))

(defun pkt-vector (pkt)
  (let ((new (make-vector (pkt-nbytes pkt))))
    (copy-data-from-packet pkt new)
    new))

(defun set-pkt-string (pkt &restv strings)
  (loop for str being the vector-elements of strings
	and start = 0 then (+ start len)
	and len = (length str)
	do (copy-data-to-packet str pkt (length str) start)))

(defsubst get-pkt () (allocate-resource 'packet))
(defsubst return-pkt (pkt) (deallocate-resource 'packet pkt))

(defun send-pkt (conn pkt &optional (opcode pkt$k_dat))
  (prog1 (send conn ':send-pkt ':pkt pkt ':opcode opcode)
	 (return-pkt pkt)))

(defun send-string (conn &restv strings)
  (with-packet (pkt)
    (lexpr-funcall 'set-pkt-string pkt strings)
    (send conn ':send-pkt ':pkt pkt ':opcode pkt$k_dat)))

(defun send-unc-pkt (conn pkt &optional pkt-number ack-number)
  (when pkt-number (setf (pkt$w_packet pkt) pkt-number))
  (when ack-number (setf (pkt$w_ack pkt) ack-number))
  (setf (pkt$b_opcode pkt) pkt$k_unc)
  (send conn ':send-pkt ':pkt pkt))

(defun may-transmit (conn)
  (chk-vms-open conn "CHAOS:MAY-TRANSMIT")
  (send conn ':may-transmit))

(defun data-available (conn)
  (chk-vms-open conn "CHAOS:DATA-AVAILABLE")
  (send conn ':data-available))

(defun get-next-pkt (conn &optional (timeout 0))
  (chk-vms-open conn "CHAOS:GET-NEXT-PKT")
  (if (eq timeout t)
      (cond ((and (memq (send conn ':state)
			'(#.conn$k_closed #.conn$k_los #.conn$k_incom))
		  (not (data-available conn)))
	     nil)
	    (t (send conn ':get-packet ':timeout 0)))
      (send conn ':get-packet ':timeout timeout)))

(defun remove-conn (conn)
  (send conn ':remove-self))

(defun close-conn (conn &optional (reason ""))
  (when (send conn ':vms-open?)
    (when (= (send conn ':state) conn$k_open)
      (send conn ':close ':reason reason))
    (remove-conn conn)))

(defun state (conn)
  (chk-vms-open conn "CHAOS:STATE")
  (cdr (assq (send conn ':state) *chaos-state-alist*)))

(fset 'conn-state (fsymeval 'state))

(defun connect (host contact-name &optional (timeout 5400))
  (let ((new nil)
	state (ok nil))
    (unwind-protect
     (progn
      (setq new (allocate-resource 'connection))
      (setq state
	    (send new ':rfc ':host host ':contact-name contact-name ':timeout timeout))
      (unless state (ferror ':chaos-timeout "RFC Timeout"))
      (when (not (= state conn$k_open))
	(case state
	  (#.conn$k_closed
	   (if (data-available new)
	       (with-packet (pkt)
		 (send new ':get-packet ':pkt pkt)
		 (case (pkt$b_opcode pkt)
		   (#.pkt$k_cls
		    (ferror ':chaos-lossage "Connection refused by ~A: ~S"
			    (find-host-name (pkt$w_src_host pkt))
			    (pkt-string pkt)))
		   (#.pkt$k_ans
		    (ferror ':chaos-lossage
			    "ANS recieved instead of OPN from ~A"
			    (find-host-name (pkt$w_src_host pkt))))))
	       (ferror ':chaos-lossage "Connection closed by foreign host")))
	  (#.conn$k_los
	   (if (data-available new)
	       (with-packet (pkt)
		 (send new ':get-packet ':pkt pkt)
		 (ferror ':chaos-lossage
			 "LOS recieved trying to connect to ~A: ~S"
			 (find-host-name (pkt$w_src_host pkt))
			 (pkt-string pkt)))))))
      (setq ok t))
     (unless ok (when new
		  (deallocate-resource 'connection new)
		  (setq new nil))))
    new))

(defun listen-conn (contact-name &optional (timeout 5400))
  (let ((new nil)
	(state nil)
	(ok nil))
    (unwind-protect
     (progn
      (setq new (allocate-resource 'connection))
      (setq state (send new ':listen
			':contact-name contact-name
			':timeout timeout))
      (unless state
	(setq state (send new ':state))
	(if (= state conn$k_lsn)
	    (ferror ':chaos-timeout "Listen Timeout")
	    (ferror ':chaos-lossage
		    "Listen error: connection state = ~S"
		    (state-sym state))))
      (setq ok t))
     (unless ok (when new
		  (deallocate-resource 'connection new)
		  (setq new nil))))
    new))
      
(defun simple (host contact-name &optional (timeout 5400))
  (using-resource (connection new)
    (or (send new ':rfc ':host host ':contact-name contact-name ':timeout timeout)
	(ferror ':chaos-timeout "RFC Timeout"))
    (when (not (= (send new ':state) conn$k_closed))
;;      (send new ':close ':carefully nil)
      (ferror ':chaos-lossage "OPN recieved instead of ANS"))
    (let ((pkt nil) (ok nil) (errstr nil))
      (unwind-protect
       (progn (setq pkt (allocate-resource 'packet))
	      (send new ':get-packet ':pkt pkt)
	      (if (= (pkt$b_opcode pkt) pkt$k_cls)
		  (setq errstr (pkt-string pkt))
		  (setq ok t)))
       (unless ok
	 (when pkt (deallocate-resource 'packet pkt))
	 (setq pkt errstr)))
      pkt)))

(defun accept (conn)
  (chk-vms-open conn "CHAOS:ACCEPT")
  (with-csb (buff)
    (let ((state (send conn ':state ':buffer buff)))
      (when (= state conn$k_rfcrcv)
	(when (not (zerop (conn$w_rcv_available buff)))
	  (with-packet (pkt)
	    (send conn ':get-packet ':pkt pkt))) ; get the RFC
	(send conn ':send-buffer ':data "" ':opcode pkt$k_opn)))))

(defun reject (conn reason)
  (chk-vms-open conn "CHAOS:REJECT")
  (let ((state (send conn ':state)))
    (if (not (= state conn$k_rfcrcv))
	(ferror ':chaos-lossage
		"Not in RFC recieved state: state = ~S" (state-sym state))
	(send conn ':close ':reason reason)
	(remove-conn conn))))

(defun answer-string (conn string)
  (chk-vms-open conn "CHAOS:ANSWER-STRING")
  (with-packet (pkt)
    (set-pkt-string pkt string)
    (answer conn pkt)))

(defun answer (conn pkt)
  (chk-vms-open conn "CHAOS:ANSWER")
  (let ((state (send conn ':state)))
    (when (not (= state conn$k_rfcrcv))
      (ferror ':chaos-lossage
	      "Connection not in RFC Recieved state, state = ~S"
	      (state-sym state)))
    (send conn ':send-pkt ':pkt pkt ':opcode pkt$k_ans)
    (remove-conn conn)))

(defun wait (conn state timeout &optional ignore)
  (chk-vms-open conn "CHAOS:WAIT")
  (let ((staten (if (numberp state) state (state-number state))))
    (send conn ':wait ':old-state staten ':timeout timeout)))

(defun finish (conn &optional ignore)
  (chk-vms-open conn "CHAOS:FINISH")
  (if (not (= (send conn ':state) conn$k_open))
      nil
      (send conn ':finish)))

(defun chaos-patch-editing-function (file)
  (list "Now" "edit" file))

;;;; connection information functions

(defun conn-foreign-address (conn)
  (send conn ':foreign-host))

(defun conn-read-pkts (conn)
  (send conn ':data-available))

(defun conn-window-available (conn)
  (with-csb (buff)
    (send conn ':state ':buffer buff)
    (conn$w_xmit_available buff)))

(globalize 'conn-foreign-address 'chaos)
(globalize 'conn-read-pkts 'chaos)
(globalize 'conn-window-available 'chaos)

(globalize 'pkt-data  'chaos)
(globalize 'pkt-data-word  'chaos)
(globalize 'set-pkt-data 'chaos)
(globalize 'set-pkt-data-word 'chaos)
(globalize 'pkt-opcode  'chaos)
(globalize 'pkt-nbytes  'chaos)
(globalize 'pkt-string  'chaos)
(globalize 'pkt-vector  'chaos)
(globalize 'set-pkt-string  'chaos)
(globalize 'get-pkt  'chaos)
(globalize 'return-pkt  'chaos)
(globalize 'send-pkt  'chaos)
(globalize 'send-string  'chaos)
(globalize 'send-unc-pkt  'chaos)
(globalize 'may-transmit  'chaos)
(globalize 'data-available 'chaos)
(globalize 'get-next-pkt 'chaos)
(globalize 'simple 'chaos)
(globalize 'connect 'chaos)
(globalize 'listen-conn 'chaos)
(globalize 'remove-conn 'chaos)
(globalize 'close-conn 'chaos)
(globalize 'accept 'chaos)
(globalize 'reject 'chaos)
(globalize 'answer-string 'chaos)
(globalize 'answer 'chaos)
(globalize 'state 'chaos)
(globalize 'conn-state 'chaos)
(globalize 'wait 'chaos)
(globalize 'finish 'chaos)
