;;; -*- Mode: Lisp; Base: 10; Package: HOST -*-

(herald host)
(eval-when (compile eval)
  (or (get 'nhosttab 'version)
      (load "nil$chaos:nhosttab")))

(defflavor host-table-part
	   ((is-open? nil)
	    (rms-file nil)
	    (file-stream nil)
	    (name "")
	    (buffer-size 0))
	   ()
  (:initable-instance-variables name buffer-size)
  (:gettable-instance-variables name buffer-size is-open?))

(defflavor host-table
	   ((nets (make-host-table-part "nets" "idx" net$k_length))
	    (name (make-host-table-part "name" "idx" nik$k_length))
	    (addr (make-host-table-part "addr" "idx" nid$k_length))
	    (dial (make-host-table-part "dial" "tab" 10))
	    (host (make-host-table-part "host" "tab" hst$k_length))
	    (rand (make-host-table-part "rand" "tab" hst$k_rand_len))
	    (net-number-list nil))
	   ())

(defun make-host-table-part (name type buf-size)
  (make-instance 'host-table-part
		 ':name (string-append "chaos$disk:[chaos.host.tables]net"
				       name
				       "2."
				       type)
		 ':buffer-size buf-size))

(defvar *the-host-tables* nil)

(defun init-host-tables (&optional &key (force nil))
  (if *the-host-tables*
      (if force
	  (progn (send *the-host-tables* ':finish)
		 (setq *the-host-tables* 'nil)
		 (init-host-tables)))
      (setq *the-host-tables* (make-instance 'host-table))))

(eval-when (compile eval load)
  (defsubst maybe-init-host-tables ()
    (unless *the-host-tables*
      (init-host-tables)))

  (defmacro def-hosttab-op (operation)
      `(progn 'compile
	      (defun ,operation (&restv args)
		(unwind-protect
		 (progn
		  (maybe-init-host-tables)
		  (lexpr-send *the-host-tables* ',(intern operation 'keyword)
			      args))
		 (and *the-host-tables*
		  (send *the-host-tables* ':close-all))))
	      (globalize ',operation 'chaos)))

  (defmacro with-buffer ((var size) &body body)
    `(si:using-resource (si:bit-buffer ,var (with-buffer-size* ,size))
			,@body))

  (defmacro fixnum-to-bits (num bits)
    `(setf (nibble ,bits 0 (min 30 (simple-bit-vector-length ,bits))) ,num))

  (defmacro with-fixnum-in-bits ((num var &optional (size 32)) &body body)
    `(si:using-resource (si:bit-buffer ,var ,size)
			(fixnum-to-bits ,num ,var)
			,@body))
  )

(defun with-buffer-size* (size)
  (cond ((typep size 'host-table-part) (*& 8 (send size ':buffer-size)))
	((typep size 'fixnum) size)
	(t 0)))

(defun bits-to-string (bits)
  (let (str-len new)
    (setq str-len
	  (loop for i from 0 below (/& (simple-bit-vector-length bits) 8)
		when (= (get-a-byte bits i) 32)
		 do (return i)))
    (setq new (make-string str-len))
    (dotimes (i str-len)
      (setf (char new i) (character (get-a-byte bits i))))
    new))

(def-hosttab-op net-number)
(def-hosttab-op host-address)
(def-hosttab-op host-name)
(def-hosttab-op host-info)

;;;; host-table-part methods

(defmethod (host-table-part :open) ()
  (if (null name) (ferror () "FOO!"))
  (if (not (null is-open?))
      "Already open"
      (progn (setq rms-file (rms:open-rms-file name))
	     (setq file-stream (rms:make-rms-stream rms-file))
	     (setq is-open? T))))

(defmethod (host-table-part :get) (&restv args)
  (if (not is-open?) (send self ':open))
  (lexpr-send file-stream ':get args))

(defmethod (host-table-part :find) (&restv args)
  (if (not is-open?) (send self ':open))
  (lexpr-send file-stream ':find args))

(defmethod (host-table-part :close) ()
  (when is-open?
    (send file-stream ':disconnect)
    (send rms-file ':close)
    (setq file-stream nil
	  rms-file nil
	  is-open? nil)))

;;;; host-table methods

(defmethod (host-table :finish) ()
  (send nets ':close)
  (send name ':close)
  (send addr ':close)
  (send dial ':close)
  (send host ':close)
  (send rand ':close)
  (setq nets nil)
  (setq name nil)
  (setq addr nil)
  (setq dial nil)
  (setq host nil)
  (setq rand nil)
  (setq net-number-list nil))


(defmethod (host-table :close-all) ()
  (send nets ':close)
  (send name ':close)
  (send addr ':close)
  (send dial ':close)
  (send host ':close)
  (send rand ':close))

(defmethod (host-table :net-number) (net-name)
  (let ((temp (assoc net-name net-number-list)))
    (if temp (cdr temp)
	(send self ':file-get-net-number net-name))))

(defmethod (host-table :file-get-net-number) (net-name &aux ok? status)
  (with-buffer (rec nets)
    (multiple-value (ok? status)
      (send nets ':get ':key net-name ':ubf rec ':no-error t))
    (unless ok?
      (ferror () "Can't find network: ~S~%~
	      VMS Error code: ~S" 
	      net-name (si:decode-vms-error-code status)))
    (let ((new (net$b_number rec)))
      (push (cons net-name new) net-number-list)
      new)))

(defmethod (host-table :host-address) (host-name &optional (net "CHAOS"))
  (let ((net-number (send self ':net-number net))
	(host-index (host-name-to-index host-name name)))
    (host-index-to-address host-index host net-number)))

(defmethod (host-table :host-name) (address &optional (net "CHAOS"))
  (let ((net-number (send self ':net-number net)))
    (with-buffer (adrec addr)
      (with-fixnum-in-bits (address adrbuf)
	(setf (nibble adrbuf 24 8) net-number)
	(send addr ':get ':key adrbuf ':ubf adrec)
	(with-buffer (hstrec host)
	  (with-fixnum-in-bits ((nid$w_host adrec) hstbuf)
	    (send host ':get ':key hstbuf ':ubf hstrec)
	    (bits-to-string (hst$t_name hstrec))))))))

(defmethod (host-table :host-info) (host-spec &optional (net "CHAOS")
					      &aux net-number index)
  (setq net-number (send self ':net-number net))
  (cond ((stringp host-spec) (setq index (host-name-to-index host-spec name)))
	((fixp host-spec) (setq index
				(host-address-to-index host-spec
						       addr
						       net-number)))
	(t (ferror () "Unknown type of host specification: ~S" host-spec)))
  (with-buffer (hstrec host)
    (with-fixnum-in-bits (index indbuf)
      (send host ':get ':key indbuf ':ubf hstrec)
      (list (bits-to-string (hst$t_name hstrec))
	    (bits-to-string (hst$t_machine hstrec))
	    (bits-to-string (hst$t_opsys hstrec))
	    (character (hst$b_serverp hstrec))
	    (get-nicknames index name)
	    (get-addrs hstrec)))))

;;;; Random host table routines

(defun host-name-to-index (name nicknames)
  (with-buffer (rec nicknames)
    (send nicknames ':get ':key name ':ubf rec ':krf 0)
    (nik$w_host rec)))

(defun host-address-to-index (address addr-file &optional (net-number 7))
  (with-buffer (rec addr-file)
    (with-fixnum-in-bits (address adbuf)
      (setf (nad$b_net adbuf) net-number)	; make it an inter-net address
      (send addr-file ':get ':key adbuf ':ubf rec ':krf 0)
      (nid$w_host rec))))

(defun host-index-to-address (index host-file net-number)
  (with-buffer (rec host-file)
    (with-fixnum-in-bits (index index-buf)
      (send host-file ':get ':key index-buf ':ubf rec)
      (loop for addrn from 0 below (hst$b_addr_cnt rec)
	    for skip from 528 by 32
	    as addr = (si:bits-subseq rec skip 32)
	    when (= (nad$b_net addr)
		    net-number)
	      do (return (nibble addr 0 24))
	    finally (ferror () "Host ~S not on network number ~S"
			    (bits-to-string (hst$t_name rec)) net-number)))))

(defun get-addrs (hstrec)
  (loop for addrn from 0 below (hst$b_addr_cnt hstrec)
	for skip from 528 by 32
	as addr = (si:bits-subseq hstrec skip 32)
	collect (list (nad$b_net addr)
		      (nibble addr 0 24))))

(defun get-nicknames (host-index nicknames &aux (read-index host-index))
  (with-fixnum-in-bits (host-index indbuf 16)
    (send nicknames ':find ':key indbuf ':krf 1))
  (with-buffer (nikbuf nicknames)
    (loop while (send nicknames ':get ':ubf nikbuf ':no-error t)
	  while (= (nik$w_host nikbuf) host-index)
	  collect (bits-to-string (nik$t_name nikbuf)))))
