;;; -*- Mode: Lisp; Package: CHAOS; Base: 10. -*-

;;; Byte stream chaos interface

(eval-when (compile eval)
  (or (get 'chaos:pktdefs 'version)
      (load "nil$chaos:pktdefs"))
  (or (get 'ci:chaos 'version)
      (load "nil$chaos:chaos")))


(defflavor chaos-basic-stream
	   ((conn nil)			; the connection!
	    )
	   (si:stream)
  :flavor-not-instantiable
  (:gettable-instance-variables conn)
  )

(defmethod (chaos-basic-stream :set-conn) (nc)
  (when conn
    (ferror nil "Attempt to change the associated connection of ~S" self))
  (setq conn nc))

;;; chaos-stream close

(defmethod (chaos-basic-stream :close) (&optional ignore)
  (close-conn conn))

(defmethod (chaos-basic-stream :foreign-host) ()
  (send conn ':foreign-host))

(defmethod (chaos-basic-stream :characters) () nil)	; assume binary

;;;; chaos-stream flavors

(defflavor chaos-input-stream-mixin
	   ((in-buffer nil)		; the current input packet
	    (in-count 0)		; bytes left in the input buffer
	    (in-pointer 0)		; pointer into input-packet
	    (buffer-get-function nil)
	    (unget-buffer nil)
	    (eof-seen nil))
	   ()
  :flavor-not-instantiable
  (:required-methods :new-input-buffer-really)
  (:included-flavors si:input-stream)
  (:documentation :mixin "Input side of a Chaosnet byte stream")
  )

;;;; chaos-stream input methods

(defmethod (chaos-input-stream-mixin :direction) () ':input)

(defmethod (chaos-input-stream-mixin :read-byte) ()
  (cond ((not (null unget-buffer))
	 (prog1 unget-buffer (setq unget-buffer nil)))
	(eof-seen nil)
	(in-buffer
	 (prog1 (funcall buffer-get-function in-buffer in-pointer)
		(setq in-pointer (1+ in-pointer))
		(setq in-count (1- in-count))
		(when (= in-count 0)
		  (return-pkt in-buffer)
		  (setq in-buffer nil))))
	(t
	 (send self ':new-input-buffer)
	 (send self ':read-byte))))


(defmethod (chaos-input-stream-mixin :new-input-buffer) ()
  (let* ((conn (send self ':conn))
	 (val (errset (multiple-value
		       (in-buffer in-count buffer-get-function)
		       (send self ':new-input-buffer-really conn))
		      nil)))
    (when (null val)
      (ferror ':io-lossage
	      "Attempt to read from connection in illegal state:~%~
	       conn = ~S, state = ~S" conn (conn-state conn)))
    (case (pkt$b_opcode in-buffer)
      (#.pkt$k_eof
       (return-pkt in-buffer)
       (setq in-buffer nil eof-seen t)
       nil)
      (#.pkt$k_cls
       (let ((errstr (pkt-string in-buffer)))
	 (return-pkt in-buffer)
	 (setq in-buffer nil eof-seen t)
	 (ferror ':io-lossage "Connection got an error: ~S" errstr)))
      (t (setq in-pointer 0)))))

(defmethod (chaos-input-stream-mixin :tyi) ()
  (send self ':read-byte))

(defmethod (chaos-input-stream-mixin :read-char) ()
  (let ((c (send self ':read-byte)))
    (if (null c) nil
	(character c))))

(defmethod (chaos-input-stream-mixin :unread-char) (c)
  (setq unget-buffer c))

(defmethod (chaos-input-stream-mixin :clear-eof) ()
  (setq eof-seen nil))

(defflavor chaos-output-stream-mixin
	   ((out-buffer nil)		; the current output packet
	    (out-count 0)		; free bytes left in the output buffer
	    (out-pointer 0)		; pointer into output buffer
	    (buffer-put-function nil)	; function to store into buffer
	    )
	   ()
  :flavor-not-instantiable
  (:required-methods :send-output-buffer :new-output-buffer)
  (:included-flavors si:output-stream)
  (:documentation :mixin "Output side of a Chaosnet byte stream")
  )

;;;; chaos-stream output methods

(defmethod (chaos-output-stream-mixin :direction) () ':output)

(defmethod (chaos-output-stream-mixin :write-byte) (c)
  (when (null out-buffer)
    (multiple-value (out-buffer out-count buffer-put-function)
      (send self ':new-output-buffer))
    (setq out-pointer 0))
  (funcall buffer-put-function out-buffer out-pointer c)
  (setq out-pointer (1+ out-pointer))
  (setq out-count (1- out-count))
  (when (zerop out-count)
    (send self ':flush-output))
  c)

(defmethod (chaos-output-stream-mixin :tyo) (c)
  (send self ':write-byte c))

(defmethod (chaos-output-stream-mixin :write-char) (c)
  (send self ':write-byte (char-code c)))

(defmethod (chaos-output-stream-mixin :flush-output) ()
  (when out-buffer
    (send self ':send-output-buffer (send self ':conn) out-buffer out-count)
    (setq out-buffer nil)))

(defmethod (chaos-output-stream-mixin :close) (&optional abort?)
  (unless (eq abort? ':abort)
    (send self ':eof))		; tell them we're done 
  (close-conn (send self ':conn)))

(defmethod (chaos-output-stream-mixin :eof) (&aux conn)
  (setq conn (send self ':conn))
  (send self ':flush-output)
  (send conn ':send-buffer ':data "" ':opcode pkt$k_eof)
  (send conn ':finish))

(defmethod (chaos-output-stream-mixin :finish) ()
  (send (send self ':conn) ':finish))

(defflavor chaos-io-stream-mixin
	   ()
	   (chaos-input-stream-mixin chaos-output-stream-mixin)
  :flavor-not-instantiable
  (:documentation :mixin "Support for bidirectional Chaosnet byte streams")
  )

(defmethod (chaos-io-stream-mixin :direction) () ':bidirectional)

;;;; Binary streamness

(defflavor chaos-binary-input-stream-mixin
	   ()
	   ()
  :flavor-not-instantiable)

(defmethod (chaos-binary-input-stream-mixin :new-input-buffer-really) (conn)
  (let ((pkt (get-next-pkt conn) ))
    (values pkt
	    (/ (pkt-nbytes pkt) 2)
	    #'(lambda (p w)
		(pkt-data-word p w)))))

(defflavor chaos-binary-output-stream-mixin
	   ()
	   ()
  :flavor-not-instantiable)


(defmethod (chaos-binary-output-stream-mixin :new-output-buffer) ()
  (values (get-pkt)
	  #.(/ pkt$k_maxdata 2)
	  #'(lambda (p w c) (set-pkt-data-word p w c))))

(defmethod (chaos-binary-output-stream-mixin :send-output-buffer) (c b count)
  (setf (pkt-nbytes b) (* count 2))	; count is in words
  (send-pkt c b #o300))			; use the word style opcode

;;;; character streams

(defflavor chaos-input-character-stream-mixin
	   ()
	   ()
  :flavor-not-instantiable)

(defmethod (chaos-input-character-stream-mixin :characters) () t)

(defmethod (chaos-input-character-stream-mixin :new-input-buffer-really) (conn)
  (let ((pkt (get-next-pkt conn)))
    (values pkt (pkt-nbytes pkt) #'(lambda (p b) (pkt-data p b)))))

(defflavor chaos-output-character-stream-mixin
	   ()
	   ()
  :flavor-not-instantiable)

(defmethod (chaos-output-character-stream-mixin :characters) () t)

(defmethod (chaos-output-character-stream-mixin :new-output-buffer) ()
  (values (get-pkt) pkt$k_maxdata #'(lambda (p b c) (set-pkt-data p b c))))

(defmethod (chaos-output-character-stream-mixin :send-output-buffer)
           (c buf count)
  (setf (pkt-nbytes buf) count)		; number of bytes
  (send-pkt c buf dat-op))		; send the packet



;;; ascii translation

(defflavor ascii-translating-input-stream-mixin
	   ()
	   ()
  :flavor-not-instantiable
  (:required-methods :read-char :unread-char))


(defmethod (ascii-translating-input-stream-mixin :read-char) (&aux ch)
  (case (char-code (setq ch (send-forward self ':read-char)))
    (#o210 #\BS)
    (#o211 #\TAB)
    (#o212 #\LINE)
    (#o214 #\FF)
    (#o215 #\CR)
    (#o207 #\RUBOUT)
    (T CH)))


(defflavor ascii-translating-output-stream-mixin
	   ((gobble-lf nil))
	   ()
  :flavor-not-instantiable
  (:required-methods :write-char))


(defmethod (ascii-translating-output-stream-mixin :write-char) (ch)
  (cond ((and gobble-lf (char-equal ch #\nl))
	 (setq gobble-lf nil)
	 ch)
	(t
	 (setq gobble-lf nil)
	 (send-forward self ':write-char
		       (case CH
			  (#\BS #.(int-char #o210))
			  (#\TAB #.(int-char #o211))
			  (#\LINE #.(int-char #o212))
			  (#\FF #.(int-char #o214))
			  (#\CR (setq gobble-lf t)
				#.(int-char #o215))
			  (#\RUBOUT #.(int-char #o207))
			  (T CH))))))

;;;; Instantiable flavors

(defflavor chaos-input-stream
	   ()
	   (chaos-binary-input-stream-mixin
	    chaos-input-stream-mixin
	    chaos-basic-stream)
  (:documentation :combination
		  "A binary network input stream"))

(defflavor chaos-output-stream
	   ()
	   (chaos-binary-output-stream-mixin
	    chaos-output-stream-mixin
	    chaos-basic-stream)
  (:documentation :combination
		  "A binary network output stream"))

(defflavor chaos-stream
	   ()
	   (chaos-binary-input-stream-mixin
	    chaos-binary-output-stream-mixin
	    chaos-io-stream-mixin
	    chaos-basic-stream)
  (:documentation :combination
		  "A bidirectional binary network stream"))

;;; character streams

(defflavor chaos-input-character-stream
	   ()
	   (chaos-input-character-stream-mixin
	    chaos-input-stream-mixin
	    chaos-basic-stream))

(defflavor chaos-output-character-stream
	   ()
	   (chaos-output-character-stream-mixin
	    chaos-output-stream-mixin
	    chaos-basic-stream))

(defflavor chaos-character-stream
	   ()
	   (chaos-input-character-stream-mixin
	    chaos-output-character-stream-mixin
	    chaos-io-stream-mixin
	    chaos-basic-stream))

;;; ascii translating streams

(defflavor chaos-input-ascii-translating-stream
	   ()
	   (ascii-translating-input-stream-mixin
	    chaos-input-character-stream))

(defflavor chaos-output-ascii-translating-stream
	   ()
	   (ascii-translating-output-stream-mixin
	    chaos-output-character-stream))

(defflavor chaos-ascii-translating-stream
	   ()
	   (ascii-translating-input-stream-mixin
	    ascii-translating-output-stream-mixin
	    chaos-character-stream))

(defun chaos:make-stream (conn &key &optional (direction ':bidirectional)
			       (characters t) (ascii-translation nil)
			       timeout	; ignored
			       type	;also ignored
			       &aux new
			       )
  timeout type
  (setq new
	(make-instance
	 (case direction
	   (:input
	    (cond (ascii-translation 'chaos-input-ascii-translating-stream)
		  (characters 'chaos-input-character-stream)
		  (t 'chaos-binary-input-stream)))
	   (:output
	    (cond (ascii-translation 'chaos-output-ascii-translating-stream)
		  (characters 'chaos-output-character-stream)
		  (t 'chaos-binary-output-stream)))
	   (:bidirectional
	    (cond (ascii-translation 'chaos-ascii-translating-stream)
		  (characters 'chaos-character-stream)
		  (t 'chaos-binary-stream))))))
  (send new ':set-conn conn)
  new)


;;;; fancy open type

(defparameter *chaos-open-options*
  '((:timeout)
    (:direction :bidirectional :input :output)
    (:ascii-translation t nil)
    (:characters t nil)))


(defun (:chaos si:open-type) (type path &restv options)
  (when (or (typep path 'string)
	    (typep path 'list))
    (setq path (pathname path)))
  (si:check-open-options type path options *chaos-open-options*)
  (cond ((not (typep path 'pathname))
	 (values nil "Bad pathname"))
	((zerop (string-length (pathname-name path)))
	 (chaos-open-lsn path options))
	(t
	 (chaos-open-rfc path options))))

(defun chaos-open-lsn (path options)
  (let ((new nil)
	(timeout (si:find-open-option ':timeout options)))
    (if (null timeout)
	(setq new (listen-conn (pathname-type path)))
	(setq new (listen-conn (pathname-type path) timeout)))
    (values (lexpr-funcall #'chaos:make-stream new options) nil)))

(defun chaos-open-rfc (path options)
  (chaos-open-rfc-1 (pathname-name path) (pathname-type path) options))

(defun open-stream (host contact &restv options)
  (chaos-open-rfc-1 host contact options))

(defun chaos-open-rfc-1 (host contact options)
  (let ((new nil)
	(timeout (si:find-open-option ':timeout options)))
    (if (null timeout)
	(setq new (connect host contact))
	(setq new (connect host contact timeout)))
    (values (lexpr-funcall #'chaos:make-stream new options) nil)))
